import stack_api.api_core as api_core
from src.comm.docker_ver import *
import typer
from pathlib import Path

from src.core.core import *

app = typer.Typer()

# Checks if local files are installed
try:
    api = api_core.API()
    initialized = api.start_check()
    assert(initialized)
except:
    try:
        import os
        os.remove(path_home+'/.config_stack')
    except:
        print(f'Stack has not been installed yet!')
        print(f'setting-up Stack...')
    api = api_core.API()
    initialized = api.init()

    if initialized:
        api.start_check()

### CLI End-points ###

@app.command("init")
def init(uri: str):
    if docker_ver():
        if uri[0] == '.':
            uri = uri[1:]
    return {'success': api.init(uri)}

@app.command("add")
def add_command(path: str, subpath: str=''):
        try:
            if len(subpath)>1:
                if subpath[-1] != '/':
                    subpath = subpath + '/'
            add(api.Initializer,[path],subpath)
            api.commit('')
            return True
        except:
            return Exception


@app.command("remove")
def remove_command(key: str, subpath: str=''):
    if len(subpath)>1:
        if subpath[-1] != '/':
            subpath = subpath + '/'
    remove(api.Initializer,[key],subpath)
    api.commit('')

# End-points
@app.command("connect")
def connect_cli(uri: str):
    try:
        api.init(uri)
        api.connect_post_cli()
        api.start_check()
        return {'success': True}
    except:
        return {'success': False}

@app.command("disconnect")
def disconnect_cli(uri: str):
    try:
        return {'success': api.disconnectDataset(uri)}
    except:
        return {'success': False}

@app.command("remove_full")
def full_remove_key_api(key: str):
    try:
        api_core.remove_full(key)
        print(f"Removed: {key}")
        api.commit('')
    except:
        print(f"Unable to remove: {key}")

@app.command("uri")
def uri_api():
    uri = api.getURI()
    print(uri)

@app.command("history")
def history_api():
	printHistory(api.Initializer)

@app.command("datasets")
def dataset_api():
    api.print_datasets()


@app.command("status")
def status_api():
    try:
        printStatus(api.Initializer)
    except:
        return Exception

@app.command("sync")
def sync_api(comment: str=''):
    # TODO: do we need to print something?
    return {'success': api.commit(comment)}

@app.command("pull")
def pull_file_api(file):
    try:
        return api.pull(file)
    except:
        return {}


@app.command("diff")
def diff_api(v1: str, v0: str, file: str=''):
    printDiff(api.Initializer, v1, v0, file)
    return True


@app.command("diff_csv")
def diff_csv(csv1: str, csv2: str):
    from csv_diff import load_csv, compare
    diff = compare(load_csv(open(csv1)), load_csv(open(csv2)))
    print(diff)
    return True


@app.command("revert")
def revert_api(version):
    assert(version != '')
    revertCommit(api.Initializer, int(version))
    api.commit('')
    return True


if __name__ == "__main__":
    app()



import sys
sys.path.append( '../../../' )
import os
import shutil
from pathlib import Path
import time
from src.comm.docker_ver import *
path_home = '/localpath/' if docker_ver() else str(Path.home())
ref_path = '/localpath/' if docker_ver() else ''

class Local(object):
	"""docstring for Storage"""
	def __init__(self):
		self.type = "local"
		self.dataset = "./"
		self.credentials = {}

	def createDataset(self,location):
		# transforms to absolute path
		
		if docker_ver():
			if location[0] == '/':
				location = ref_path + location[1:]
			else:
				location = ref_path + location
		else:
			if location[0] == '~':
				if len(location) > 1:
					location = path_home+location[1:]
				else:
					location = path_home
			if location[0] != '/':
				location = str(os.path.abspath(location))

		#print('Initializing dataset at '+location)
		if location[-1] != '/':
			location = location + '/'
		if not os.path.exists(location):
			os.makedirs(location)
		self.dataset = location
		return True

	def checkIfEmpty(self,path):
		if os.path.exists(path):
			dir = os.listdir(path)
			return (len(dir) == 0)
		else:
			return True

	def addFile(self,filepath,target_name='',subpath=''):
		if filepath[-1] == '/':
			for root,dirs,files in os.walk(filepath):
				for file in files:
					if not os.path.exists(os.path.dirname(self.dataset+subpath)):
						os.makedirs(os.path.dirname(self.dataset+subpath))
					head_tail = os.path.split(os.path.join(root,file))
					shutil.copyfile(os.path.join(root,file), self.dataset+subpath+head_tail[1])	
				self.resetBuffer()
		else:
			if not os.path.exists(os.path.dirname(self.dataset+subpath)):
				os.makedirs(os.path.dirname(self.dataset+subpath))
			head_tail = os.path.split(filepath)
			shutil.copyfile(filepath, self.dataset+subpath+head_tail[1])
		return True

	def addFolder(self,folderpath):
		if not os.path.exists(self.dataset + folderpath):
			os.makedirs(self.dataset + folderpath)
		return True

	def addFileFromBinary(self,filename,f):
		return self.addFileFromBinaryGlobal(self.dataset+filename,f)

	def addFileFromBinaryGlobal(self,filename,f):
		if not os.path.exists(os.path.dirname(filename)):
			os.makedirs(os.path.dirname(filename))
		binary_file = open(filename, "wb")
		binary_file.write(f.read())
		binary_file.close()
		return True

	def removeFile(self,filename):
		return self.removeFileGlobal(self.dataset+filename)

	def removeFileGlobal(self,filename):
		if os.path.exists(filename):
			if filename[-1] == '/':
				shutil.rmtree(filename)
			else:
				os.remove(filename)
		return True

	def loadFile(self,filename):
		return self.loadFileGlobal(self.dataset+filename)

	def loadFileBytes(self,filename,bi,bf):
		return self.loadFileGlobal(self.dataset+filename,bi,bf)

	def loadFileGlobal(self,filename):
		return open(filename,'rb')

	def loadFileGlobalBytes(self,filename,bi,bf):
		f = open(filename, "rb")
		f.seek(bi)
		data = f.read(bf-bi)
		return data

	def loadFileMetadata(self,filename):
		path = self.dataset+filename
		metadata = {
			'key' : self.dataset+filename,
			'date_loaded' : time.strftime("%m/%d/%Y, %H:%M:%S", time.gmtime(os.path.getmtime(path))),
			'date_added' : time.strftime("%m/%d/%Y, %H:%M:%S", time.gmtime(os.path.getctime(path))),
			'last_modified' : time.strftime("%m/%d/%Y, %H:%M:%S", time.gmtime(os.path.getmtime(path))),
		}
		return metadata

	def loadDataset(self):
		file_m = []
		for root, dirs, files in os.walk(self.dataset, topdown=False):
			for name in files:
				path = os.path.join(root, name)
				if path[-1] != '/':
					metadata = {
						'key' : path,
						'last_modified' : time.strftime("%m/%d/%Y, %H:%M:%S", time.gmtime(os.path.getmtime(path)))
					}
					file_m.append(metadata)
		return file_m

	
	def loadDatasetList(self):
		return self.loadListInPath(self.dataset)

	def loadListInPath(self,path):
		keys = []
		last_m = []

		for root, dirs, files in os.walk(path, topdown=False):
			for name in files:
				keys.append(os.path.join(root, name))
				last_m.append(time.strftime("%m/%d/%Y, %H:%M:%S", time.gmtime(os.path.getmtime(os.path.join(root, name)))))
		return keys, last_m

	def listFilesinDataset(self):
		files = []
		for root, dirs, files in os.walk(self.dataset, topdown=False):
			for name in files:
				path = os.path.join(root, name)
				print(path)
		return True

	def listEverythinginDataset(self):
		for root, dirs, files in os.walk(self.dataset, topdown=False):
			for name in files:
				path = os.path.join(root, name)
				print(path.name)
		return True

	def listFilesinPath(self,dir_path):
		for root, dirs, files in os.walk(self.dataset + dir_path, topdown=False):
			for name in files:
				path = os.path.join(root, name)
				print(path)
		return True

	def listEverythinginPath(self,dir_path):
		for root, dirs, files in os.walk(self.dataset + dir_path, topdown=False):
			for name in files:
				path = os.path.join(root, name)
				print(path)
		return True

	def copyFile(self,filepath,full_target_name):
		return self.copyFileGlobal(self.dataset + filepath, self.dataset + full_target_name)

	def copyFileGlobal(self,filepath,full_target_name):
		head_tail = os.path.split(filepath)
		os.makedirs(os.path.dirname(full_target_name), exist_ok=True)
		if filepath[-1] != '/':
			shutil.copyfile(filepath, full_target_name)
		return True

	def resetBuffer(self):
		pass

def main():
	storage = Local()
	
	print('create dataset')
	storage.createDataset('/Users/bernardo/test_data/')
	
	print('add folder')
	storage.addFolder('.stack')
	print('add subfolder')
	storage.addFolder('subtree')
	print('list in dataset')
	storage.listFilesinDataset()
	print('list in path')
	storage.listFilesinPath('')
	
	print('adds a file')
	storage.addFile('../../tests/image.png')
	print(storage.loadFileMetadata('image.png'))

	storage.copyFile('image.png',storage.dataset+'/copies/image4.png')
	storage.listFilesinPath('')

if __name__ == '__main__':
	main()


from operator import le
import sys
sys.path.append( '..' )
from src.core.init import Initializer
from src.core.core import *
from src.storage.classes.s3 import S3Bucket
from src.storage.classes.gcs import GCSBucket
from src.storage.classes.local import Local
from pathlib import Path
import pickle

from src.comm.docker_ver import *
path_home = '/localpath/' if docker_ver() else str(Path.home())

class API(object):
    """docstring for CLI"""
    def __init__(self, reset=False):
        super(API, self).__init__()
        self.key_bin = None
        self.Initializer = None
        if not Path(path_home+'/datasets.stack').exists():
            self.set_datasets({})
        if reset:
            self.Initializer = None
        
        if not Path(path_home+'/.config_stack').exists():
            self.set_config({})
        else:
            config = self.get_config()
            try:
                self.storage_name = config['storage']
                self.dataset_name = config['dataset']
                ctype = config['type']
            except:
                config = self.set_config({})
                ctype = ''

            if ctype == 'local':
                cloud = Local()
                cloud.createDataset(config['dataset'])
                self.Initializer = Initializer(cloud)
                self.dataset_name = self.Initializer.storage.dataset
                self.storage_name = self.Initializer.storage.dataset
            elif ctype == 's3':
                cloud = S3Bucket(config['bucket'])
                cloud.connectBucket()
                cloud.createDataset(config['dataset'])
                self.Initializer = Initializer(cloud)
            elif ctype == 'gcs':
                cloud = GCSBucket(config['bucket'])
                cloud.connectBucket()
                cloud.createDataset(config['dataset'])
                self.Initializer = Initializer(cloud)
            else:
                self.Initializer = None
        
    def init(self, storage = None):
        # builds a config file       
        if Path(path_home+'/.config_stack').exists():
            config = {} 
            if storage == None:
                return False
            if 's3' in storage.lower():
                bucket_data = storage.split("/")[1:]
                config['bucket'] = bucket_data[1]
                config['dataset'] = ''.join(bucket_data[2:])+'/'
                config['type'] = 's3'
            elif 'gs' in storage.lower():
                bucket_data = storage.split("/")[1:]
                config['bucket'] = bucket_data[1]
                config['dataset'] = ''.join(bucket_data[2:])+'/'
                config['type'] = 'gcs'
            else:
                config['dataset'] = storage
                config['type'] = 'local'
            config['storage'] = storage
            # stores the config file

            print('Initializing dataset in ' + storage.lower())
            self.set_config(config)
            # creates dataset
            return True
        else:
            print('Creating config file')
            config = {'storage': ''}
            self.set_config(config)
            return False

    def set_gs_key(self, file):
        self.key_bin = file
        return True

    def get_config(self):
        file2 = open(path_home+'/.config_stack', 'rb')
        config = pickle.load(file2)
        file2.close()
        return config

    def set_config(self, config):
        file = open(path_home+'/.config_stack', 'wb')
        pickle.dump(config,file)
        file.close()

    def get_datasets(self):
        file1 = open(path_home+'/datasets.stack', 'rb')
        datasets = pickle.load(file1)
        file1.close()
        return datasets

    def set_datasets(self, datasets):
        file1 = open(path_home+'/datasets.stack', 'wb')
        pickle.dump(datasets,file1)
        file1.close()

    def print_datasets(self):
        datasets = self.get_datasets()

        if len(datasets.keys()) > 0:
            for s in datasets.keys():
                print('-- '+datasets[s]['name']+' in '+datasets[s]['storage'])
            print('')
            print('run \'stackcli connect [dataset_uri]\' to connect one of these datasets')
        else:
            print('no datasets to show')
        return True
    
    def connect_post_web(self, name='My Dataset', keys={}):
        
        print('connecting to '+name)
        config = self.get_config()

        self.storage_name = config['storage']
        self.dataset_name = config['dataset']
        
        try: 
            datasets = self.get_datasets()
        except:
            datasets = {}

        is_not_there = True
        for s in datasets.keys():
            if datasets[s]['storage'] == self.storage_name:
                is_not_there = False

        if is_not_there:
            datasets[self.storage_name] = {'storage': self.storage_name, 'name': name, 'type': config['type']}
            self.set_datasets(datasets)

        if config['type'] == 'local':
            cloud = Local()
            cloud.createDataset(config['dataset'])
            self.Initializer = Initializer(cloud)
            self.dataset_name = self.Initializer.storage.dataset
            self.storage_name = self.Initializer.storage.dataset
        elif config['type'] == 's3':
            cloud = S3Bucket(config['bucket'])
            cloud.connect_bucket_api(keys)
            cloud.createDataset(config['dataset'])
            self.Initializer = Initializer(cloud)
        elif config['type'] == 'gcs':
            cloud = GCSBucket(config['bucket'])
            cloud.connect_bucket_api(self.key_bin)
            cloud.createDataset(config['dataset'])
            self.Initializer = Initializer(cloud)
        else:
            self.Initializer = None
        return True

    def connect_post_cli(self):
        config = self.get_config()

        self.storage_name = config['storage']
        self.dataset_name = config['dataset']
        
        try: 
            datasets = self.get_datasets()
        except:
            datasets = {}

        is_not_there = True
        for s in datasets.keys():
            if datasets[s]['storage'] == self.storage_name:
                is_not_there = False

        if is_not_there:
            name = input('please give a name to your dataset: ')
            datasets[self.storage_name] = {'storage': self.storage_name, 'name': name, 'type': config['type']}
            self.set_datasets(datasets)

        if config['type'] == 'local':
            cloud = Local()
            cloud.createDataset(config['dataset'])
            self.Initializer = Initializer(cloud)
            self.dataset_name = self.Initializer.storage.dataset
            self.storage_name = self.Initializer.storage.dataset
        elif config['type'] == 's3':
            cloud = S3Bucket(config['bucket'])
            cloud.connectBucket()
            cloud.createDataset(config['dataset'])
            self.Initializer = Initializer(cloud)
        elif config['type'] == 'gcs':
            cloud = GCSBucket(config['bucket'])
            cloud.connectBucket()
            cloud.createDataset(config['dataset'])
            self.Initializer = Initializer(cloud)
        else:
            self.Initializer = None
        return True

    def connect_post_api(self, name='My Dataset'):
        config = self.get_config()

        self.storage_name = config['storage']
        self.dataset_name = config['dataset']
        
        try: 
            datasets = self.get_datasets()
        except:
            datasets = {}

        is_not_there = True
        for s in datasets.keys():
            if datasets[s]['storage'] == self.storage_name:
                is_not_there = False

        if is_not_there:
            datasets[self.storage_name] = {'storage': self.storage_name, 'name': name, 'type': config['type']}
            self.set_datasets(datasets)

        if config['type'] == 'local':
            cloud = Local()
            cloud.createDataset(config['dataset'])
            self.Initializer = Initializer(cloud)
            self.dataset_name = self.Initializer.storage.dataset
            self.storage_name = self.Initializer.storage.dataset
        elif config['type'] == 's3':
            cloud = S3Bucket(config['bucket'])
            cloud.connectBucket()
            cloud.createDataset(config['dataset'])
            self.Initializer = Initializer(cloud)
        elif config['type'] == 'gcs':
            cloud = GCSBucket(config['bucket'])
            cloud.connectBucket()
            cloud.createDataset(config['dataset'])
            self.Initializer = Initializer(cloud)
        else:
            self.Initializer = None
        return True

    def disconnectDataset(self, storage=''):
        assert(len(storage) > 0)
        datasets = self.get_datasets()
        print('disconnecting from ' + storage)

        for s in datasets.keys():
            if datasets[s]['storage'] == storage:
                del datasets[s]
                self.set_datasets(datasets)
                return True
        return False

    def connectDataset(self, storage=None):
        # checks if another dataset exists
        # builds a config file
        if Path(path_home+'/.config_stack').exists():
            config = self.get_config()
            
            if config['dataset'] != storage:
                print('connecting to dataset')

            config['dataset'] = storage

            if config['type'] == 'local':
                cloud = Local()
                cloud.createDataset(config['dataset'])
                config['dataset'] = cloud.dataset
                self.Initializer = Initializer(cloud)
            elif config['type'] == 's3':
                cloud = S3Bucket(config['bucket'])
                cloud.connectBucket()
                cloud.createDataset(config['dataset'])
                self.Initializer = Initializer(cloud)
            elif config['type'] == 'gcs':
                cloud = GCSBucket(config['bucket'])
                cloud.connectBucket()
                cloud.createDataset(config['dataset'])
                self.Initializer = Initializer(cloud)

            # stores the config file
            self.set_config(config)

            # creates dataset
            return True
        else:
            return False

    
    def upload_file_binary(self, filename='', binary=''):
        assert(filename != '')
        assert(binary != '')
        add_from_binary(self.Initializer, filename, binary)
        return True

    def start_check(self):
        try:
            return self.Initializer.start_check()
        except:
            return False

    def getURI(self):
        return {'storage': self.storage_name, 'dataset': self.dataset_name}

    def add(self, path, subpath=''):
        if len(subpath)>1:
            if subpath[-1] != '/':
                subpath = subpath + '/'
        add(self.Initializer,[path],subpath)
        return True

    def pull_all(self,version = 'current'):
        print('downloading files from last commit')
        metapath = self.Initializer.prefix_meta + 'current.json'
        current = json.load(self.Initializer.storage.loadFileGlobal(metapath))
        pull(self.Initializer, current['keys'], version)
        return True

    def pull(self, file, version = 'current'):
        if file == '.' or file == 'all':
            print('downloading all contents')
            return self.pull_all(version)
        else:
            if not self.Initializer.storage.dataset in file:
                file = self.Initializer.storage.dataset + file
                print('downloading ' + file)
            pull(self.Initializer, [file], version)
        return True

    def pull_file(self, file, version = 'current'):
        
        if not self.Initializer.storage.dataset in file:
            file = self.Initializer.storage.dataset + file
        
        if version == 'current':
            # saves each file
            for key in files:
                binary = self.Initializer.storage.loadFileGlobal(key)
        else:
            gtfo = False
            # finds the commit of interest
            metapath = self.Initializer.prefix_meta+'history.json'
            history = self.Initializer.load(self.Initializer.storage.loadFileGlobal(metapath))
            for key in files:
                if key[-1] == '/':
                    print('Do not pull directories')
                    return False
                for i in range(len(history),int(version)-1,-1):
                    for commit in history[str(i)]['commits']:
                        # reads each file version
                        cmit = json.load(self.Initializer.storage.loadFileGlobal(commit))
                        if str(cmit['version']) == version and cmit['key'] == key:
                            if cmit['type'] != 'remove':
                                key = self.Initializer.prefix_diffs + key + '/' + str(cmit['version']).zfill(10)
                                binary = self.Initializer.storage.loadFileGlobal(key).read()
                                self.Initializer.storage.resetBuffer()
                            gtfo = True
                        if gtfo:
                            break
                if gtfo:
                    gtfo = Falsfe
                    break

        self.Initializer.storage.resetBuffer()
        file_array = {binary: binary, key: file}

        return file_array

    def commits_version(self, version = 2, l = 5, page = 0):
        assert(int(l) > 0)
        assert(int(version) >= 0)
        assert(int(page) >= 0)

        metapath = self.Initializer.prefix_meta+'history.json'
        history = json.load(self.Initializer.storage.loadFileGlobal(metapath))
        self.Initializer.storage.resetBuffer()

        response = {}
        idx = 0

        i_p = int(page)*int(l)
        i_f = min((int(page)+1)*int(l),len(history[str(int(version))]['commits']))

        # goes over the commits
        for i in range(i_p, i_f):
            # reads each file version
            commit = history[str(int(version))]['commits'][i]
            cmit = json.load(self.Initializer.storage.loadFileGlobal(commit))
            self.Initializer.storage.resetBuffer()
            response[idx] = {'key': cmit['key'], 'source': cmit['source'], 'date': history[str(int(version))]['date'], 'comment': cmit['comment']}
            idx = idx + 1

        return {'commits': response, 'len': len(history[str(int(version))]['commits'])}

    def key_versions(self, key = '', l = 5, page = 0):
        assert(int(l) > 0)
        assert(int(page) >= 0)
        key_hist = get_key_history(self.Initializer, self.Initializer.storage.dataset + key)
        response = {}
        i_p = len(key_hist) - int(page)*int(l)
        i_f = max(len(key_hist) - int(l)*(int(page)+1),0)

        idx = 0

        # goes over the commits
        for i in range(i_p, i_f, -1):
            # reads each file version
            response[idx] = key_hist[str(i)]
            idx = idx+1

        return {'commits': response, 'len': len(key_hist)}

    def status(self):
        metapath = self.Initializer.prefix_meta+'current.json'
        return json.load(self.Initializer.storage.loadFileGlobal(metapath))

    def remove(self, key, subpath=''):
        if len(subpath)>1:
            if subpath[-1] != '/':
                subpath = subpath + '/'
        remove(self.Initializer,[key],subpath)
        return True

    def remove_commit(self, version = '-1'):
        assert(int(version) >= 0)
        
        metapath = self.Initializer.prefix_meta+'history.json'
        history = json.load(self.Initializer.storage.loadFileGlobal(metapath))
        self.Initializer.storage.resetBuffer()

        for i in range(len(history[str(int(version))]['commits'])):
            commit = history[str(int(version))]['commits'][i]
            cmit = json.load(self.Initializer.storage.loadFileGlobal(commit))
            self.Initializer.storage.resetBuffer()
            removeGlobal(self.Initializer, [cmit['diff']])

        return True

    def remove_key_diff(self, key, version):
        remove_diff(self.Initializer,self.Initializer.storage.dataset+key,int(version))
        return True

    def remove_key_full(self, key, version):
        remove_full(self.Initializer,self.Initializer.storage.dataset+key)
        return True

    def commit(self, comment=''):
        commit(self.Initializer, comment)
        print('sync done!')
        return True

    def loadCommitMetadata(self, commit_file):
        try:
            return json.load(self.Initializer.storage.loadFileGlobal(commit_file))
        except:
            return {}

    def check_if_setup(self):
        self.Initializer.setupDataset()
        print('setup complete!')
        return True

    def load_file_binary(self, file, version='current'):
        print('THIS IS THE VERSION ' + version)
        if version=='current':
            print('loading ' + file + '...')
            return self.Initializer.storage.loadFile(file)
        elif int(version) >= 1:
            print('loading ' + file + ' version '+ version +'...')
            path = self.Initializer.prefix_diffs + self.Initializer.storage.dataset + file + '/' + str(int(version)).zfill(10)
            return self.Initializer.storage.loadFileGlobal(path)
        else:
            assert(False)

    def load_file_binary_bytes(self, file, bi, bf):
        print('loading '+file+'...')
        return self.Initializer.storage.loadFileBytes(file,bi,bf)

    def reset(self):
        self.Initializer.removeSetup()
        self.Initializer.setupDataset()
        print('setup complete!')
        return True

    def revert(self, version=0):
        assert(version != '')
        revertCommit(self.Initializer, int(version))
        commit(self.Initializer, 'reverted to version ' + str(version))

    def revert_file(self, key, version):
        try: 
            if self.Initializer.storage.dataset in key:
                revertFile(self.Initializer, key, int(version))
            else:
                revertFile(self.Initializer, self.Initializer.storage.dataset+key, int(version))
            self.Initializer.storage.resetBuffer()

            return True
        except:
            return False

    def history(self):
        metapath = self.Initializer.prefix_meta+'history.json'
        return json.load(self.Initializer.storage.loadFileGlobal(metapath))

    def lastNcommits(self, n = 0):
        metapath = self.Initializer.prefix_meta+'history.json'
        history = json.load(self.Initializer.storage.loadFileGlobal(metapath))
        self.Initializer.storage.resetBuffer()

        response = {}
        idx = 0

        # goes over history
        for i in range(len(history),0,-1):
            for commit in history[str(i)]['commits']:
                # reads each file version
                cmit = json.load(self.Initializer.storage.loadFileGlobal(commit))
                response[idx] = {'source': cmit['source'], 'date': history[str(i)]['date'], 'comment': cmit['comment']}

                idx += 1
                if int(idx) >= int(n):
                    break
            if int(idx) >= int(n):
                    break

        return response

    def diff(self, v1, v0, file=''):
        printDiff(self.Initializer, v1, v0, file)
        return True


import sys
sys.path.append( '..' )
import stack_api.api_core as api_core
from fastapi import FastAPI, File, UploadFile, Response, Form
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from fastapi.encoders import jsonable_encoder
from starlette.responses import StreamingResponse
from pathlib import Path

# API Definition
app = FastAPI()
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"], 
)

from src.comm.docker_ver import *
path_home = '/localpath/' if docker_ver() else str(Path.home())

# checks if local files are installed
try:
    api = api_core.API()
    initialized = api.start_check()
    assert(initialized)
except:
    try:
        import os
        os.remove(path_home+'/.config_stack')
    except:
        print('no config file')
    api = api_core.API()
    initilized = api.init()

    if initilized:
        api.start_check()

# End-points
@app.get("/init")
async def init(uri='', name='My Dataset'):
    try:
        api.init(uri)
        api.connect_post_api(name)
        api.start_check()
        return {'success': True}
    except:
        return {'success': False}

@app.post("/init_web/")
async def init_web(data: dict):
    try:
        api.init(data['uri'])
        api.connect_post_web(data['name'], data)
        api.start_check()
        json_compatible_item_data = jsonable_encoder(True)
        return JSONResponse(content=json_compatible_item_data)
    except:
        json_compatible_item_data = jsonable_encoder(False)
        return JSONResponse(content=json_compatible_item_data)

@app.post("/init_gskey/")
async def init_gskey(file: UploadFile = File(description="A file read as UploadFile")):
    try:
        api.set_gs_key(file.file)
        return {'success': True}
    except:
        return {'success': False}

@app.get("/directories")
async def directories():
    import os
    from pathlib import Path
    print(path_home)
    print(str(os.path.abspath('.')))
    return {'success': True}

@app.get("/disconnect")
async def disconnect_api(uri=''):
    try: 
        return {'success': api.disconnectDataset(uri)}
    except:
        return {'success': False}

@app.get("/get_datasets")
async def get_datasets_api():
    try:
        return api.get_datasets()
    except:
        return {}

@app.get("/uri")
async def uri_api():
    try:
        return api.getURI()
    except:
        return {}

@app.get("/history")
async def history_api():
    try:
        return api.history()
    except:
        return {}

@app.post("/add_file/")
async def add_file_api(file: UploadFile = File(description="A file read as UploadFile")):
    try:
        api.upload_file_binary(file.filename, file.file)
        api.commit('')
        return {'success': True}
    except:
        return {'success': False}

@app.get("/commits_version")
async def commits_version_api(version=1,l=5, page=0):
    try:
        return api.commits_version(version, l, page)
    except:
        return {}

@app.get("/key_versions")
async def key_versions_api(key='',l=5, page=0):
    try:
        return api.key_versions(key, l, page)
    except:
        return {}

@app.get("/last_n_commits")
async def last_n_commits_api(n=5):
    try:
        return api.lastNcommits(n)
    except:
        return {}

@app.get("/last_commits_from_hist_api")
async def last_commits_from_hist_api(n=1):
    try:
        return api.getHistoryCommits(n)
    except:
        return {}

@app.get("/status")
async def status_api():
    try:
        return api.status()
    except:
        return {}

@app.get("/commit_req")
async def commit_api(comment=''):
    return {'success': api.commit(comment)}

@app.get("/get_commit_metadata")
async def get_commit_meta_api(commit):
    try:
        return api.loadCommitMetadata(commit)
    except:
        return {}

@app.get("/pull_file_api")
async def pull_file_api(file, version='current'):
    try:
        return StreamingResponse(api.load_file_binary(file, version), media_type="image/png")
    except:
        return Response(content='')

@app.get("/remove_key")
async def remove_key_api(key):
    try:
        api.remove(key)
        api.commit('')
        return {'sucess': True}
    except:
        return {'sucess': False}

@app.get("/remove_commit")
async def remove_commit(version):
    try:
        api.remove_commit(version)
        api.commit('')
        return {'sucess': True}
    except:
        return {'sucess': False}

@app.get("/full_remove_key")
async def full_remove_key_api(key):
    try:
        api.remove_full(key)
        api.commit('')
        return {'sucess': True}
    except:
        return {'sucess': False}

@app.get("/remove_key_version")
async def remove_key_diff_api(key, version=-1):
    try:
        api.remove_key_diff(key, version)
        api.commit('')
        return {'sucess': True}
    except:
        return {'sucess': False}

@app.get("/revert_key_version")
async def revert_key_version_api(key, version=-1):
    try:
        api.revert_file(key, version)
        api.commit('reverted file ' + key)
        return {'sucess': True}
    except:
        return {'sucess': False}

@app.get("/revert")
async def revert_api(version=0):
    try:
        api.revert(version)
        api.commit('reverted file ' + key)
        return {'success': True}
    except:
        return {'success': False}


import socket
import string
import json
import io
from datetime import datetime
import time
from pathlib import Path

from src.comm.docker_ver import *
path_home = '/localpath/' if docker_ver() else str(Path.home())

class Initializer(object):
	"""docstring for Initializer"""
	def __init__(self, storage ,user=None):
		super(Initializer, self).__init__()
		self.storage = storage
		self.dataset = storage.loadDataset()
		self.user = user
		
		# prefixes
		if self.storage.type == 'local':
			if not docker_ver():
				surrogate_dataset = self.storage.dataset.replace(self.storage.dataset, '/localpath/')
			else:
				surrogate_dataset = self.storage.dataset

			self.prefix_meta = path_home + '/.stack' +  surrogate_dataset
			self.prefix_curr = path_home + '/.stack' +  surrogate_dataset + 'current/'
			self.prefix_commit = path_home + '/.stack' + surrogate_dataset + 'commits/'
			self.prefix_history = path_home + '/.stack' + surrogate_dataset + 'history/'
			self.prefix_diffs = path_home + '/.stack' + surrogate_dataset + 'diffs/'
		else:
			self.prefix_meta = '.stack/'+self.storage.dataset
			self.prefix_curr = '.stack/'+self.storage.dataset+'current/'
			self.prefix_commit = '.stack/'+self.storage.dataset+'commits/'
			self.prefix_history = '.stack/'+self.storage.dataset+'history/'
			self.prefix_diffs = '.stack/'+self.storage.dataset+'diffs/'

		self.start_check()

	def defineSchema(self):
		# defines the dataset schema for exporting and visualizing
		return True

	def removeSetup(self):
		if self.storage.type == 'local':
			self.storage.removeFileGlobal(path_home+'/.stack/')
		else:
			self.storage.removeFileGlobal('.stack/')
		return True

	def start_check(self):
		# checks if the dataset exists
		if not self.verify_setup():
			self.setupDataset()
		return True

	def verify_setup(self):
		# verifies if a setup .stack/ directory exists
		if self.storage.checkIfEmpty(self.prefix_meta):
			return False

		if self.storage.checkIfEmpty(self.prefix_curr):
			return False

		if self.storage.checkIfEmpty(self.prefix_diffs):
			return False

		if self.storage.checkIfEmpty(self.prefix_commit):
			return False

		return True

	def setupDataset(self):
		# performs all key operations
		self.copyCurrent()
		self.setupDiffs()
		self.setupCommits()
		self.setupHistory()
		
		return True

	def copyCurrent(self):
		# backup of the current version to keep track of changes
		if self.storage.checkIfEmpty(self.prefix_curr):
			for file in self.dataset:
				self.storage.copyFileGlobal(file['key'],self.prefix_curr+file['key'])
		else:
			self.storage.removeFile(self.prefix_curr)
			self.dataset = self.storage.loadDataset()
			for file in self.dataset:
				self.storage.copyFileGlobal(file['key'],self.prefix_curr+file['key'])	
		self.storage.resetBuffer()

		# saves the metadata of the backup
		print('loading dataset list')
		metapath = self.prefix_meta + 'current.json'
		keys, lm = self.storage.loadDatasetList()
		current = {'keys': keys, 'lm': lm}
		self.storage.addFileFromBinaryGlobal(metapath,io.BytesIO(json.dumps(current).encode('ascii')))
		self.storage.resetBuffer()
		return True

	def copyCurrentCommit(self, toadd, todelete):
		# backup of the current version to keep track of changes
		if self.storage.checkIfEmpty(self.prefix_curr):
			for file in toadd:
				self.storage.copyFileGlobal(file,self.prefix_curr+file)
			for file in todelete:
				self.storage.removeFileGlobal(self.prefix_curr+file)
		else:
			for file in toadd:
				self.storage.copyFileGlobal(file,self.prefix_curr+file)
			for file in todelete:
				self.storage.removeFileGlobal(self.prefix_curr+file)
		self.storage.resetBuffer()

		# saves the metadata of the backup
		metapath = self.prefix_meta + 'current.json'
		print('loading dataset list')
		keys, lm = self.storage.loadDatasetList()
		current = {'keys': keys, 'lm': lm}
		self.storage.addFileFromBinaryGlobal(metapath,io.BytesIO(json.dumps(current).encode('ascii')))
		self.storage.resetBuffer()
		return True

	def setupDiffs(self):
		# adds all files, creates a diff for each
		# located in .stack/meta/diffs/abcde/DIFF_ID

		# stores the diffs of the first commit
		for file in self.dataset:
			# TODO: let's name the diffs with just numbering for now
			diff = self.prefix_diffs + file['key'] + '/' + str(1).zfill(10)
			self.storage.copyFileGlobal(file['key'],diff)		
		return True

	def setupHistory(self):
		# meta-data
		# located in .stack/meta/history.json
		""" 
			json = {
				0 : ['c1.json','c2.json'],
				1 : ['c3.json'],
			}
		"""
		commits = []
		for file in self.dataset:
			commitpath = self.prefix_commit + file['key'] + '/' + str(1).zfill(10)
			commits.append(commitpath)

		time = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
		history = {1: {'commits': commits, 'date': time}}
		metapath = self.prefix_meta+'history.json'
		self.storage.addFileFromBinaryGlobal(metapath,io.BytesIO(json.dumps(history).encode('ascii')))

		return True

	def setupCommits(self):
		# meta-data for each commit
		# located in .stack/meta/commits/abcde/COMMIT_ID
		""" 
			json = {
				'file' : 'img1_jpg',
				'id:   : 'abcdefg'
				'diff' : 'dff1',
				'type' : 'add',
				'date' : [...],
				'source' : '0.0.0.0',
				'comment' : '',
			}
		"""

		# checks who is the user
		idx = 0
		if self.user == None:
			username = socket.gethostbyname(socket.gethostname())
		else:
			username = self.user.username

		# stores the medadata of the first commit
		for file in self.dataset:
			# gets the address of the diff
			# TODO: let's name the diffs with just numbering for now
			diff = self.prefix_diffs + file['key'] + '/' + str(1).zfill(10)

			commit = {
				'key'	: file['key'],
				'diff'	: diff,
				'type'	: 'add',
				'version' : 1,
				'date'	: file['last_modified'],
				'source': username,
				'comment' : 'added '+ file['key'],
			}

			history = {}
			history[1] = commit
			commitpath = self.prefix_commit + file['key'] + '/' + str(1).zfill(10)
			self.storage.addFileFromBinaryGlobal(commitpath,io.BytesIO(json.dumps(commit).encode('ascii')))
			
			self.storage.resetBuffer()
			
			histpath = self.prefix_history + file['key'] + '/history.json'
			self.storage.addFileFromBinaryGlobal(histpath,io.BytesIO(json.dumps(history).encode('ascii')))
			self.storage.resetBuffer()

		self.storage.resetBuffer()
		
		return True

	def getLatestDiffNumber(self, key):
		# checks all the diffs
		diff_path, _ = self.storage.loadListInPath(self.prefix_diffs + key + '/')
		
		# gets the list in number
		diff_path = [int(x.replace(self.prefix_diffs + key + '/','')) for x in diff_path]
		
		if len(diff_path):
			return max(diff_path)
		else:
			return 0

	def getListtoCompare(self):
		# get new files
		new_files, new_lm = self.storage.loadDatasetList()
		metapath = self.prefix_meta + 'current.json'
		current = json.load(self.storage.loadFileGlobal(metapath))

		return new_files, new_lm, current['keys'], current['lm']

def main():
	import sys
	sys.path.append( '../..' )

	from src.storage.classes.s3 import S3Bucket
	
	cloud = S3Bucket('stacktest123')
	cloud.connectBucket()
	cloud.createDataset('dataset1/')
	print('dataset created')

	init = Initializer(cloud)
	print('init created')

	init.copyCurrent()
	print('current version copied')

	init.setupDiffs()
	print('first diffs added')

	init.setupCommits()
	print('first commits added')

	init.setupHistory()
	print('first change added')

	init.storage.copyFile('image.png','image2.png')
	
	nf,_,of,_ = init.getListtoCompare()
	print('list of new files')
	print(nf)

	print('list of old files')
	print(of)

if __name__ == '__main__':
	main()



# from src.storage import Storage
import os
import logging
from google.cloud import storage
import maskpass
import io
from pathlib import Path
import sys
sys.path.append( '../../../' )
from src.comm.docker_ver import *
path_home = '/localpath/' if docker_ver() else str(Path.home())

class GCSBucket(object):
	"""docstring for Storage"""
	def __init__(self, BUCKET_NAME):
		# self = Cloud().__init__()
		self.type = "gcs"
		self.dataset = ""
		self.BUCKET_NAME = BUCKET_NAME
		self.credentials = {}
		self.client = None
		self.bucket = None

	def connectBucket(self):
		# creates a client 
		key_path = path_home+'/.gs_key'
		os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = key_path
		try:		
			self.client = storage.Client()
		except:
			# in case the keys are not in the computer
			print('1) Create a service account here: https://console.cloud.google.com/projectselector/iam-admin/serviceaccounts/create')
			print('2) Give it access to Cloud Storage')
			print('3) Go to your new service account/keys')
			print('4) Create a new key in .json format and save it as ~/.gs_key')
			return False

		# checks if the bucket exists
		print('Connecting to your bucket...')
		buckets = self.client.list_buckets()
		found = False
		for bucket in buckets:
			if bucket.name == self.BUCKET_NAME:
				found = True
				break

		# connects to bucket or finds another one
		if found:
			self.bucket = self.client.get_bucket(self.BUCKET_NAME)
		else:			
			print('Could not find your desired bucket...')
			print('Do you want to creat a bucket with the name '+self.BUCKET_NAME+"?")
			yn = input("[Y/n]: ")
			if yn == "y" or yn == "Y":
				self.bucket = self.client.create_bucket(self.BUCKET_NAME)
			else:
				print('pick another bucket from this list:')
				for bucket in buckets:
					print("-" + bucket.name)
				self.BUCKET_NAME = input("Enter another bucket name: ")
				return self.connectBucket()
		return True

	def connect_bucket_api(self,binary):
		# reads the gs_key
		print('creating key file')
		key_path = path_home+'/.gs_key'
		binary_file = open(key_path, "wb")
		binary_file.write(binary.read())
		binary_file.close()

		# sets env variable for google cloud
		print('exporting key')	
		os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = key_path

		print('connecting to client')
		try:
			self.client = storage.Client()
		except:
			return False

		# checks if the bucket exists
		print('Connecting to your bucket...')
		buckets = self.client.list_buckets()
		for bucket in buckets:
			if bucket.name == self.BUCKET_NAME:
				self.bucket = self.client.get_bucket(self.BUCKET_NAME)
				return True
		return False

	def createDataset(self,location):
		# assigns location
		if location[-1] != '/':
			location = location + '/'
		self.dataset = location
		return True

	def addFile(self,filepath,target_name='',subpath=''):
		# adding new file
		if target_name == '':
			target_name = os.path.basename(filepath)
		if filepath[-1] == '/':
			for root,dirs,files in os.walk(filepath):
				for file in files:
					blob = self.bucket.blob(self.dataset+subpath+os.path.basename(filepath))
					blob.upload_from_filename(os.path.join(root,file))	
			self.resetBuffer()
		else:
			blob = self.bucket.blob(self.dataset+subpath+os.path.basename(filepath))
			blob.upload_from_filename(filepath)
		return True

	def addFileFromBinary(self,filename,f):
		return self.addFileFromBinaryGlobal(self.dataset+filename,f)

	def addFileFromBinaryGlobal(self,filename,f):
		blob = self.bucket.blob(filename)
		blob.upload_from_file(file_obj=f, rewind=True)
		return True

	def removeFile(self,filename):
		return self.removeFileGlobal(self.dataset+filename)

	def removeFileGlobal(self,filename):
		# deletes files in path
		if filename[-1] == '/':
			for blob in self.client.list_blobs(self.BUCKET_NAME, prefix=filename, delimiter=None):
				blob.delete()
			self.resetBuffer()
		else:
			blob = self.bucket.blob(filename)
			blob.delete()
		return True

	def loadFile(self,filename):
		return self.loadFileGlobal(self.dataset+filename)

	def loadFileBytes(self,filename,bi,bf):
		return self.loadFileGlobalBytes(self.dataset+filename,bi,bf)

	def loadFileGlobal(self,filename):
		blob = self.bucket.blob(filename)
		file_obj = blob.download_as_string()
		return io.BytesIO(file_obj)

	def loadFileGlobalBytes(self,filename,bi,bf):
		blob = self.bucket.blob(filename)
		file_obj = blob.download_as_string(start=bi,end=bf)
		return io.BytesIO(file_obj)	

	def loadFileMetadata(self,filename,debug=False):
		blob = self.bucket.get_blob(self.dataset + filename)
		if blob.name[-1] != '/':
			metadata = {
				'key' : self.dataset + filename,
				'date_loaded' : blob.time_created.strftime("%m/%d/%Y, %H:%M:%S"),
				'date_added' : blob.time_created.strftime("%m/%d/%Y, %H:%M:%S"),
				'last_modified' : blob.updated.strftime("%m/%d/%Y, %H:%M:%S"),
				'HostId' : blob.id,
			}
			return metadata

	def listFilesinPath(self,path):
		blobs = self.client.list_blobs(self.BUCKET_NAME, prefix=path, delimiter=None)
		for blob in blobs:
			print(f'-- {blob.name}')
		return True

	def loadDataset(self):
		files = []
		blobs = self.client.list_blobs(self.BUCKET_NAME, prefix=self.dataset, delimiter=None)
		# labels each file
		for blob in blobs:
			if blob.name[-1] != '/':
				metadata = {
					'key' : blob.name,
					'last_modified' : blob.updated.strftime("%m/%d/%Y, %H:%M:%S"),
				}
				files.append(metadata)
		return files

	def loadDatasetList(self):
		# loads all the metadata associated with the dataset
		return self.loadListInPath(self.dataset)

	def loadListInPath(self,path):
		# loads all the metadata associated within a path
		keys = []
		last_m = []

		blobs = self.client.list_blobs(self.BUCKET_NAME, prefix=path, delimiter=None)

		# loads each file
		for blob in blobs:
			if blob.name[-1] != '/':
				keys.append(blob.name)
				last_m.append(blob.updated.strftime("%m/%d/%Y, %H:%M:%S"))
			
		return keys, last_m

	def checkIfEmpty(self,path):
		# checks the number of files in a directory
		blobs = self.client.list_blobs(self.BUCKET_NAME, prefix=path, delimiter=None)
		idx = 0
		for b in blobs:
			idx = idx + 1
		return (idx == 0)

	def listFilesinDataset(self):
		# lists all the files in the main path of the dataset
		return self.listFilesinPath(self.dataset)

	def copyFile(self,filepath,full_target_name):
		# adding new file
		return self.copyFileGlobal(self.dataset+filepath,self.dataset+full_target_name)

	def copyFileGlobal(self,filepath,full_target_name):
		source_blob = self.bucket.blob(filepath)
		blob_copy = self.bucket.copy_blob(source_blob, self.bucket, full_target_name)
		return True

	def resetBuffer(self):
		return True

def main():
	# connects a bucket
	bucket = 'gs://stack123'
	cloud = GCSBucket('stack123')
	cloud.connectBucket()

	# connects to a dataset in the bucket
	cloud.createDataset('dataset1/')

	# list the files in the bucket
	cloud.listFilesinPath('')
	cloud.listFilesinDataset()

	# addes on file
	cloud.addFile('../../tests/image.png','image.png')
	cloud.addFile('../../tests/image.png','base/image.png',subpath='base/')
	cloud.resetBuffer()

	# opens one file
	body = cloud.loadFile('image.png')
	print(body)

	meta = cloud.loadFileMetadata('image.png')
	print(meta)

	# adds two file from binaries
	cloud.addFileFromBinary('extra/image2.png',open("../../tests/image.png", "rb"))
	cloud.addFileFromBinary('extra/image3.png',cloud.loadFile('base/image.png'))
	cloud.resetBuffer()

	# removes a file
	cloud.removeFile('extra/image2.png')

	# loads the dataset in memory
	dataset = cloud.loadDataset()
	print(cloud.loadDataset())

	cloud.listFilesinDataset()
	
	cloud.copyFileGlobal(cloud.dataset+'extra/image3.png',cloud.dataset+'extra/image4.png')
	cloud.listFilesinPath('')
	print('success')

if __name__ == '__main__':
	main()